# This file is part of the Motelab batch scripts
# Copyright (c) 2005 Bret Hull, Kyle Jamieson, all rights reserved.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA

#
# ml_ping: pings, optionally erases, and optionally connects to each
# mote in the testbed.
#
# Options:
#   --erase: erase each mote
#   --connect: test-connect to each mote
#   

use strict;
use Getopt::Long;
use DBI;
use threads;
use threads::shared;

use Socket;

use libml;

#
# Parse command line options
#
my %options=();

GetOptions(\%options, "user=s", "host=s", "port=s", "password=s",
           "erase", "connect");
ml_default_values(%options);

#
# connect to db
# 

my $dsn = "DBI:mysql:auth:$options{'host'}:$options{'port'}";
my $dbh = DBI->connect($dsn, $options{'user'}, $options{'password'});
die "Unable to connect to database $options{'host'}" unless $dbh;
print "Connected to $options{'host'}:$options{'port'} as $options{'user'}\n";

my $roomQuery = "SELECT moteid, roomlocation FROM motes WHERE moteid != 0";
my $roomStatement = $dbh->prepare($roomQuery) or 
  die "Couldn't prepare query '$roomQuery': $DBI::errstr\n";
$roomStatement->execute();
my %location;

while (my ($moteid, $location) = $roomStatement->fetchrow_array()) {
  $location{$moteid} = $location;
}

my @threadlist = ();

my $sth = $dbh->prepare("select moteid, ip_addr from motes where active=1");
$sth->execute() or die;

my @results;
my $maxid = 0;

while (my ($id,$ip) = $sth->fetchrow_array) {
  $results[$id] = &share([]);
  push @threadlist, threads->new(\&doPing, $ip, $id);
  $maxid = $id if ($id > $maxid);
}

$_->join for (@threadlist);

print "\n";

for my $arrref (@results) {
  my $hr = $arrref->[0];
  print "Note $arrref->[1] ($location{$arrref->[1]}) BAD PING\n" 
    if (exists $hr->{'ping'} and not $hr->{'ping'});

  print "Note $arrref->[1] ($location{$arrref->[1]}) CAN'T ERASE\n" 
    if (exists $hr->{'erase'} and not $hr->{'erase'});

  print "Note $arrref->[1] ($location{$arrref->[1]}) CAN'T CONNECT\n" 
    if (exists $hr->{'connect'} and not $hr->{'connect'});
}

print scalar @threadlist . " motes.\n";

sub doPing ($) {
  my ($ip,$id) = @_;

  $results[$id]->[0] = &share({});
  $results[$id]->[1] = $id;

  sleep int rand 10;
  my $p = `ping -q -c 5 -i 2 -w 10 $ip`;
  $p =~ /(\d+) packets transmitted, (\d+) received/;
  my $ratio = $2/$1;

  if ($ratio < 1) {
    $results[$id]->[0]->{'ping'} = 0;
  } else {
    $results[$id]->[0]->{'ping'} = 1;
  }

  if ($ratio > 0.5 and $options{'erase'}) {
    $p = `uisp -dprog=stk500 -dhost=$ip -dpart=ATmega128 --erase &>/dev/null`;
    my $exitCode = $? >> 8;
    if ($exitCode == 0) {
      $results[$id]->[0]->{'erase'} = 1;
    } else {
      $results[$id]->[0]->{'erase'} = 0;
    }
  }

  if ($ratio > 0.5 and $options{'connect'}) {
    $results[$id]->[0]->{'connect'} = test_connect($ip);
  }
  
  if ($results[$id]->[0]->{'ping'} eq 0) {
    print "!";
  } elsif ($results[$id]->[0]->{'erase'} eq 0) {
    print "~";
  } elsif ($results[$id]->[0]->{'connect'} eq 0) {
    print "@";
  } else {
    print "+";
  }
}

sub test_connect($) {
  my $host = shift;
  my $iaddr = gethostbyname('localhost');
  my $proto = getprotobyname('tcp');
  my $port = 10002;
  my $paddr = sockaddr_in(0, $iaddr);
  my $hisiaddr = inet_aton($host);
  my $hispaddr = sockaddr_in($port, $hisiaddr);

  socket(SOCKET, PF_INET, SOCK_STREAM, $proto) || return 0;
  connect(SOCKET, $hispaddr)          || return 0;
  close(SOCKET);
  return 1;
}
