# This file is part of the Motelab batch scripts
# Copyright (c) 2005 Bret Hull, Kyle Jamieson, all rights reserved.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA

#
# ml_sched: schedules jobs to run on the testbed.
#
# Options:
#   --duration=[d]: set the duration of the job, overriding the
#     duration given in the config file.
#   --start=[time/date string]: start the job at the given time and
#     date.  Default is next available time.
#   --config=[CONFIGFILE]: use the given configuration file.
#

package Cents;

use strict;
use Getopt::Long;
use DBI;
use Cwd;
use Time::Local;
use Date::Parse;

use libml;

#
# Parse command line options
#
my %options=();

GetOptions(\%options, "user=s", "host=s", "password=s", "host=s",
"duration=s", "noschedule", "help", "start=s", "config=s",
"prefix=s"); 
default_values();

my $nogo = 0;

if($options{'help'}) {
  $nogo = 1;
} elsif (!$options{'user'}) {
  $nogo = 1;
  printf "ml_sched: must specify --user\n";
} 

#
# if no-go, tell user what to do
#

if ($nogo) {
  print 'Usage: ml_sched [options]

Configuration options: 
  --user=<user>            (default: env var ML_DB_USER)
  --password=<psswd>       (default: env var ML_DB_PASSWORD)
  --host=<motelab db host> (default: limey.lcs.mit.edu)
  --duration=<minutes>     Length in minutes of an individual expt
  --start=next|now|<date>       When to start this run of expts (default: now)
      12/21/93 17:05
      21 Dec 93 17:05:20
      21/dec/93 17:05
      21 dec 17:05
      21-dec 17:05
      21/dec 17:05
  --config=<file>          (default: DEFAULT.cfg)
  --prefix=<prefix>     

Actions:
  --help                   Print this message.

Modifiers:
  --noschedule             Do everything except for actually scheduling the
                           job.
';
  exit;
}

#
# connect to db
# 

my $dsn = "DBI:mysql:auth:$options{'host'}:$options{'port'}";
my $dbh = DBI->connect($dsn, $options{'user'}, $options{'password'});
die "Unable to connect to database $options{'host'}" unless $dbh;
print "Connected to $options{'host'}:$options{'port'} as $options{'user'}\n";

#
# lookup user info
# 
my $ml_user_id = getUserID($options{'user'}, $dbh) or 
                 die "invalid username: $options{'user'}\n";


#
# Read in compiler options 
#
my @compiler_options;
my @compiler_values;
my @compiler_abbr;

readCompilerOptions($options{"config"});

# 
# Obtain ids of jobs to schedule
#
printf "******** GENERATING: job list\n";
my @jobids = getJobs($options{'prefix'},0,());

#
# Check for conflicts
#
printf "******** CHECKING: for conflicts\n";
my $duration = $options{'duration'};
if ($duration < 1) {
  die "Duration must be specified\n";
}
my $num_jobs = scalar(@jobids);
my $total_dur = $num_jobs * ($duration+1);
printf("num jobs: $num_jobs expt duration: $total_dur minutes\n");
if(checkConflicts($options{'start'},$total_dur)) {
  die "Unable to schedule: conflicts exist\n";
} else {
  printf "No conflicts\n";
}

#
# Schedule jobs
# 
printf "******** SCHEDULING: all jobs\n";
if(!$options{'noschedule'}) {
  schedule($options{"start"},$duration,@jobids);
}

printf "******** DONE: success\n"; 
exit(1);

#
# subroutines
# 

sub readCompilerOptions {
  my ($config) = @_;
  open COMPILER_OPTIONS, $config or die "go: can't open compiler_options\n";
  while (<COMPILER_OPTIONS>) {
    my $opt;
    my $val;
    my $abbr;
    if (/(\w+):(\w+)=(\S+)/) {
      $opt = $1;
      $val = $3;
      $abbr = $2;
    } elsif (/(\w+)=(\S+)/) {
      $opt = $1;
      $val = $2;
      $abbr = "NA";
    } else {
      die "ml_update: invalid CFG\n";
    }
    if(defined $ENV{$1}) {
      $val = $ENV{$1};
    }
  
    if( $opt eq "DURATION") {
      if($options{'duration'} eq "") {
        $options{'duration'} = $val;
      }
    } elsif($opt eq "APPDIR") {
      if($options{'prefix'} eq "") {
        $options{'prefix'} = $abbr;
      }
    } elsif(!($opt eq "MOTELAB_CLASS_FILES")) {
#      printf("ml_update: opt=$opt abbr=$abbr val=$val\n");
      push @compiler_options, $opt;
      push @compiler_values, $val;
      push @compiler_abbr, $abbr;
    }
  }
  close COMPILER_OPTIONS;
}

sub getJobs() {
  my($name,$i,@jobs) = @_;
  if(@compiler_options[$i]) {
    my $abbr = @compiler_abbr[$i];
    my @values = split /\,/,@compiler_values[$i];
    my $value = @values[0];
    if (scalar(@values) == 1 && $abbr eq "NA" ) {
      @jobs = getJobs($name,$i+1,@jobs);
    } else {      
      foreach $value (@values) {
        @jobs = getJobs("$name$abbr$value",$i+1,@jobs);
      }
    }
  } else {
    #chop $name;
    my $jobid = getJobID($name,$ml_user_id);
    if($jobid or die "unknown job $name\n") {
      printf("attempting to schedule $name\n");
      push @jobs, $jobid;
    }
  }
  return @jobs;
}

sub getStartTime($) {
  my $start = shift;

  if ($start eq "now") {
    return time();
  } elsif ($start eq "next") {
    my $sth = $dbh->prepare("SELECT UNIX_TIMESTAMP(end) " .
                            "FROM auth.jobschedule " .
                            "WHERE state=0 or state=1 or state=5 " . 
                            "ORDER BY end desc limit 1");
    $sth->execute();
    my $next = $sth->fetchrow_array() + 60;
    if($next < time()) {
      $next = time();
    }
    return $next;
  } else {
    die "Cannot schedule in the past" if str2time($start) <= time();
    my @starttime = strptime($start);
    my @curtime = localtime;

    #
    # default start date values to current date if not specified by the
    # user
    #
    for (my $i = 0; $i < @starttime; $i++) {
      $starttime[$i] = $curtime[$i] if not defined $starttime[$i];
    }
    return timelocal(@starttime);
  }
} 

sub checkConflicts($$) {
  my ($start,$duration) = @_;
  $start = getStartTime($start);
  my $end = $start + ($duration*60);
  my $query = "SELECT jobschedule.id " .
              "FROM jobschedule, jobs " .
              "WHERE jobs.id=jobschedule.jobid " .
              "AND ((start <= FROM_UNIXTIME($start) " . 
                    "AND end >= FROM_UNIXTIME($start)) " .
                   "OR (start <= FROM_UNIXTIME($end) " .
                        "AND start > FROM_UNIXTIME($start)) " .
                   "OR (end <= FROM_UNIXTIME($end) " .
                        "AND end > FROM_UNIXTIME($start)) )" . 
              "AND (state = 0 OR state = 1)";
  my $sth = $dbh->prepare($query);
  $sth->execute() or die $DBI::errstr;
  if($sth->rows() > 0) {
    print "Conflicting jobs\n";
    printf("\n%-20s %-20s %-5s %-20s\n","start", "end", "job #","prefix");
    printf("%-20s %-20s %-5s %-20s\n","-----", "---", "-----","------");
    while((my $id = $sth->fetchrow_array())) {
      printScheduledJob($id);
    }
  }
  return $sth->rows();
}

sub schedule ($$$) {
  my ($start,$job_duration,@jobs) = @_;
  my $job;
  my $start = getStartTime($start);
  my $end = $start + ($job_duration*60);
  printf("\n%-20s %-20s %-5s %-20s\n", "start", "end","job #","prefix");
  printf("%-20s %-20s %-5s %-20s\n", "-----", "---","-----","------");
  foreach $job (@jobs) {
    my $query = "INSERT INTO jobschedule " . 
                "(jobid, start, end,jobdaemon) VALUES " .
                "($job,FROM_UNIXTIME($start),FROM_UNIXTIME($end), " .
                "\"/var/www/html/testbed/util/cents-daemon.pl\")";
    $dbh->do($query) or die $DBI::errstr;
    my $sth = $dbh->prepare("SELECT id from jobschedule " .
                            "order by id desc limit 1");
    $sth->execute();
    my $id = $sth->fetchrow_array();
    printScheduledJob($id);
    $start = $end + 60;
    $end = $start + ($job_duration*60);
  }
}

sub printScheduledJob($) {
  my ($id) = @_;
  my $sth = $dbh->prepare( "select start, end, name " .
                        "from jobschedule, jobs where " .
                        "jobschedule.id=$id and jobs.id=jobschedule.jobid");
  $sth->execute();

  my @job = $sth->fetchrow_array();
  printf("%-20s %-20s %-5s %-32s\n", @job[0], @job[1], $id, 
         substr(@job[2],0,32));  
}

sub getJobID ($$) {
  my ($name,$userid) = @_;
  my $sth = $dbh->prepare("SELECT jobs.id FROM jobs WHERE " . 
            "jobs.name=\"$name\" AND jobs.userid=$userid");
  $sth->execute();
  my @row;
  if((@row=$sth->fetchrow_array())) {
    return $row[0];
  }
}

sub default_values {
  ml_default_values(%options);
  $options{"start"} = "next" if not exists $options{"start"};
  $options{"config"} = "$ENV{CENTS_HOME}/expt/DEFAULT.cfg" if not exists $options{"config"}; 
}
