# This file is part of the Motelab batch scripts
# Copyright (c) 2005 Bret Hull, Kyle Jamieson, all rights reserved.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA

#
# ml_update: update the testbed with job information.
#
# Options:
#   --config=[CONFIGFILE.cfg]: use the job configuration file 
#     CONFIGFILE.cfg.
#

package Cents;

use strict;
use Getopt::Long;
use DBI;
use Cwd;

use libml;

#
# Parse command line options
#
my %options=();

GetOptions(\%options, "user=s", "host=s", "password=s", "host=s",
"nocompile", "onlycompile", "notransfer", "config=s", "app=s","help",
"prefix=s", "scpuser=s", "motes=s", "verbose");

default_values();

my $nogo = 0;

if($options{'help'}) {
  $nogo = 1;
} elsif($options{'onlycompile'}) {
  $nogo = 0;
} elsif (!$options{'user'}) {
  $nogo = 1;
  printf "ml_update: must specify --user\n";
}
#} elsif (!$options{'password'}) {
#  $nogo= 1;
#  printf "ml_update: must specify --password\n";
#}

#
# if no-go, tell user what to do
#

if ($nogo) {
  print 'Usage: ml_update [options]

Configuration options: 
  --user=<user>            (default: env var ML_DB_USER)
  --password=<psswd>       (default: env var ML_DB_PASSWORD)
  --host=<motelab db host> (default: limey.lcs.mit.edu)
  --config=<file>          (default: DEFAULT.cfg)
  --app=<appdir>           (default: CENTS_HOME/src/apps/Expt)
  --prefix=<prefix>        
  --scpuser=<scpuser>      (default: env var ML_SCP_USER)
  --scphost=<scphost>      (default: env var ML_SCP_HOST) 
Actions:
  --help                   Print this message.

Modifiers:
  --nocompile              Do everything except compiling of binaries;
                           assumed to already exist.
  --onlycompile            Just compile the mica2 binaries.
  --notransfer             Do not scp any files to the server.
';
  exit;
}


#
# connect to db
# 

my $dsn = "DBI:mysql:auth:$options{'host'}:$options{'port'}";
my $dbh = DBI->connect($dsn, $options{'user'}, $options{'password'});
die "Unable to connect to database $options{'host'}" unless $dbh;
print "Connected to $options{'host'}:$options{'port'} as $options{'user'}\n";

#
# lookup user info
# 
my $ml_exec_dir = getExecDir($options{'user'}, $dbh);
my $ml_user_id  = getUserID($options{'user'}, $dbh);

#
# Read in compiler options 
#
my @ml_class_files;
my @compiler_options;
my @compiler_values;
my @compiler_abbr;
my @motes;

readCompilerOptions($options{"config"});

#
# Determine mote set to program
#
my @motes = getActiveMotes($dbh);
my @platforms = getMotePlatforms($dbh, @motes);

# 
# Compile binary for each expt parameter
#
chdir $options{"app"};
my @bins;
my @names;

compile($options{"prefix"},0,());
if($options{'onlycompile'}) { exit(0);}

#
# Update DB tables
#
printf("******** UPDATING: file and job tables\n");
updateDBTables();

#
# Transfer files to server 
#
printf "******** TRANSFERRING: files to Motelab\n";
if(!$options{'notransfer'}) {
  transferFiles($ml_exec_dir, \@bins, \@ml_class_files);
}

printf "******** DONE: success\n";
exit(1);

#
# subroutines
# 

sub readCompilerOptions {
  my ($config) = @_;
  open COMPILER_OPTIONS, $config or die "go: can't open compiler_options\n";
  while (<COMPILER_OPTIONS>) {
    my $opt;
    my $val;
    my $abbr;
    if (/(\w+):(\w+)=(\S+)/) {
      $opt = $1;
      $val = $3;
      $abbr = $2;
    } elsif (/(\w+)=(\S+)/) {
      $opt = $1;
      $val = $2;
      $abbr = "NA";
    } else {
      die "ml_update: invalid CFG\n";
    }
    if(defined $ENV{$1}) {
      $val = $ENV{$1};
    }
   
    if($opt eq "MOTELAB_CLASS_FILES") {
      @ml_class_files = split /\,/,$val;
    } elsif ($opt eq "APPDIR") {
      print "APPDIR=$val\n";
      $options{'app'} = "$ENV{'CENTS_HOME'}/src/apps/$val";
      if($options{'prefix'} eq "" && !($abbr eq "NA")) {
        $options{'prefix'} = $abbr;
      }
    } elsif ($opt eq "MOTES") {
      $options{'motes'} = $val;
    } else {
#      printf("ml_update: opt=$opt abbr=$abbr val=$val\n");
      push @compiler_options, $opt;
      push @compiler_values, $val;
      push @compiler_abbr, $abbr;
    }
  }
  close COMPILER_OPTIONS;
}

sub compile() {
  my($name,$i,@opts) = @_;
  my $opt = @compiler_options[$i];
  if($opt) {
    my $abbr = @compiler_abbr[$i];
    my @values = split /\,/,@compiler_values[$i];
    my $value = @values[0];
    if (scalar(@values) == 1 && $abbr eq "NA") {
      #print "$opt=$value\n";
      push @opts, "$opt=$value";
      compile("$name",$i+1,@opts);
      pop @opts;
    } else {      
      foreach $value (@values) {
        #print "$opt=$value\n";
        push @opts, "$opt=$value";
        compile("$name$abbr$value",$i+1,@opts);
        pop @opts;
      }
    }
  } else {
    #chop $name;
    push @names,$name;
    foreach my $platform (@platforms) {
      my $binname = "$name\_$platform.exe";
      push @bins,"$platform/$binname";
      foreach my $opt (@opts) {
        print "$opt\n";
      }
      printf("\n******** COMPILING $binname\n");
      if(!defined $options{'nocompile'}) {
        if ($options{'verbose'}) {
          system "@opts make $platform";
        } else {
          system "@opts make $platform > /dev/null";
        }

        if(($? >> 8) == 0) {
          system "mv build/$platform/main.exe build/$platform/$binname";
        } else {
          die "Build FAILED\n";
        }
      }
    }
  }
}

sub updateJobFiles($$$$) {
  my ($jobid,$fileids,$classids,$dbh) = @_;
 
  my $sth2 = $dbh->prepare("DELETE FROM jobfiles WHERE jobid=$jobid");
  $sth2->execute();
  
  my $mote;
  foreach $mote (@motes) {
    my $platform = getNodePlatform($dbh,$mote);
    my $fileid = $fileids->{"$platform"};
           
    $sth2 = $dbh->prepare ("INSERT INTO jobfiles (jobid,fileid,moteid) " .
                             "VALUES ($jobid,$fileid,$mote)");
    $sth2->execute();
  }
  
  my $class;
  foreach $class (@$classids) {
    $sth2 = $dbh->prepare ("INSERT INTO jobfiles (jobid,fileid,moteid) " .
                            "VALUES ($jobid,$class,0)");
    $sth2->execute();
  } 
}

sub getJobID {
  my ($name,$userid) = @_;
  my $sth = $dbh->prepare("SELECT jobs.id FROM jobs WHERE " . 
            "jobs.name=\"$name\" AND jobs.userid=$userid");
  $sth->execute();
  my @row;
  if((@row=$sth->fetchrow_array())) {
    return $row[0];
  }
}

sub addJob ($$$$) {
  my ($name,$fileids,$userid,$dbh) = @_;
  my $jobid = getJobID($name,$userid);
  my $prog = constructMoteProgram($fileids,$dbh);
  if($jobid) {
    my $sth = $dbh->prepare("UPDATE jobs SET moteprogram=$prog WHERE " .
                           "id=$jobid");
    $sth->execute();
  } else {
    my $sth = $dbh->prepare("INSERT INTO jobs
       (name,owner,disttype,currentpanel, moteprogram,userid) " .
                        "VALUES (\"$name\", " .
                        "\"$options{'user'}\", " .
                      "\"single\", \"namepanel\",$prog,$ml_user_id)");
    $sth->execute();
    $jobid = getJobID($name,$userid);
  }
  print "JOB ID=$jobid ($name)\n";
  return $jobid;
}

sub constructMoteProgram($$) {
  my ($fileids,$dbh) = @_;
  my $output;
  my $firsttime = 1;
  my $mote;
  foreach $mote (@motes) {
    my $platform = getNodePlatform($dbh,$mote);
    my $fileid = $fileids->{"$platform"};
    if($firsttime) {
      $firsttime=0;
      $output = "\"($mote, $fileid)";
    } else {
      $output = $output . "|($mote, $fileid)";
    }
  }
  return $output . "\"";
}

sub getFileID {
  my @row;
  my ($name,$userid) = @_;
  my $sth = $dbh->prepare("SELECT id FROM files WHERE " . 
            "files.name=\"$name\" AND " .
            "files.userid=$userid");
  $sth->execute();
  if((@row=$sth->fetchrow_array())) {
    return $row[0];
  }
}

sub addFile {
  my ($binname, $type, $userid) = @_;
  my $fileid = getFileID($binname,$userid);
  if(!$fileid) {
    my $sth = $dbh->prepare("INSERT INTO files (path,user,name,type,userid) " .
                        "VALUES (\"$ml_exec_dir\/$binname\", " .
                        "\"$options{'user'}\", " .
                        "\"$binname\", \"$type\",$userid)");
                      $sth->execute();
  
    $fileid = getFileID($binname,$userid);
  }
  print "id=$fileid ($binname)\n";
  return $fileid;
}

sub getClassFiles {
  my $godir=cwd();
  my $class_loc = "$ENV{CENTS_HOME}/src/pc/build/classes/cents/pkt";
  chdir $class_loc;
  my @classes = (split /\n/, `ls -1 *.class`);
  chdir $godir;
  return @classes;
}

sub transferFiles($$) {
  my ($ml_exec_dir, $bins, $ml_class_files) = @_;
  my $bin_loc="$options{'app'}/build";
  my $class_loc = "$ENV{CENTS_HOME}/src/pc/build/classes/cents/pkt";
  my $cmd = "scp";
  
  foreach my $bin (@$bins) { $cmd = "$cmd $bin_loc/$bin"; }
  foreach my $class (@$ml_class_files) { $cmd = "$cmd $class_loc/$class"; }

  $cmd = "$cmd $options{'scpuser'}\@$options{'scphost'}:$ml_exec_dir";

  #printf "cmd=$cmd\n";
  system($cmd);
}

sub getMotePlatforms($$) {
  my ($dbh,@motes) = @_;
  my (@platforms, %platforms_unique);

  foreach my $mote (@motes) {
    my $platform = getNodePlatform($dbh,$mote);
    if($platform) {
      push(@platforms, $platform) unless $platforms_unique{$platform}++;
    } else {
      die "unknown platform for mote $mote\n";
    }
  }
  print "found platforms: ";
  foreach my $platform (@platforms) {
    print "$platform ";
  }
  print "\n";

  return @platforms;
}

sub getNodePlatform($) {
  my ($dbh,$mote) = @_;
  my $sth = $dbh->prepare("SELECT platform FROM motes WHERE moteid=$mote ");
  $sth->execute() or die "unable to get mote platforms\n";
  return $sth->fetchrow_array();
}

sub getActiveMotes($) {
  my @motes;
  if(not exists $options{'motes'} || $options{'motes'} eq " ") {
    my $sth = $dbh->prepare("SELECT moteid FROM motes WHERE active=1 " .
                         "ORDER BY moteid");
    $sth->execute();
    @motes = map $_->[0], @{$sth->fetchall_arrayref};
  } else {
    if ($options{'motes'} =~ /^(\d+)$/) {
      push @motes, $options{'mote'};
    } elsif ($options{'motes'} =~ /^(\d+)-(\d+)$/) {
      for (my $id = $1; $id <= $2; $id++) {
        push @motes, $id;
      }
    } elsif ($options{'motes'} =~ /^\d+,(\d+)+/) {
      my @ids = split /,/, $options{'motes'};
      foreach my $id (@ids) {
        push @motes, $id;
      }
    } else {
      die "Bad argument to --motes: $options{'motes'}";
    }
  }
  

  my $size = scalar(@motes);
  print "motes($size): ";
  foreach my $mote (@motes) {
    print "$mote ";
  }
  print "\n";
  return @motes;
}

sub updateDBTables() {
  #
  # Add class files to files table
  #
  my @classids;
  foreach my $class (@ml_class_files) {
    my $id = addFile($class, "class", $ml_user_id); 
    push @classids, $id;
  }

  #
  # Add binary to files & jobs table of motelab
  # 
  my $name;
  foreach $name (@names) {
    my %fileids;
    foreach my $platform (@platforms) {
      $fileids{"$platform"} = addFile("$name\_$platform.exe", 
                                      "program", $ml_user_id);
    }
    my $jobid = addJob($name,\%fileids, $ml_user_id,$dbh);
    updateJobFiles($jobid, \%fileids, \@classids, $dbh); 
  }
}

#
# Defaults some options to certain values if they are not specified 
# on the command line.
#

sub default_values {
  ml_default_values(%options);
  $options{"config"} = "$ENV{CENTS_HOME}/expt/DEFAULT.cfg" if not exists $options{"config"}; 
  $options{"app"} = "$ENV{CENTS_HOME}/src/apps/Expt" if not exists $options{"app"}; 
  $options{"scpuser"} = $ENV{"ML_SCP_USER"} if not exists $options{"ML_SCP_USER"};
  if(exists $ENV{"ML_SCP_HOST"} && not exists $options{"scphost"}) {
    $options{"scphost"} = $ENV{"ML_SCP_HOST"};
  } else {
    $options{"scphost"} = $options{"host"};
  }
}
